@@ -1,4 +1,4 @@
-# -*- coding: utf-8 -*-
+# -* coding: utf-8 -*-
 """
     flaskext.openid
     ~~~~~~~~~~~~~~~
@@ -141,11 +141,18 @@ class SessionWrapper(object):
 
 class RegLookup(object):
 
-    def __init__(self, resp):
+    def __init__(self, resp, extensions):
         sreg_resp = SRegResponse.fromSuccessResponse(resp)
         self.sreg = sreg_resp and sreg_resp.data or {}
         self.ax_resp = ax.FetchResponse.fromSuccessResponse(resp) or {}
 
+        # Process the OpenID response with the OpenIDResponse class provided
+        self.ext = {}
+        for extension in extensions:
+            ext_name = getattr(extension, 'ns_alias', extension.__name__)
+            self.ext[ext_name] = \
+                extension.fromSuccessResponse(resp)
+
     def get(self, name, default=None):
         assert name in ALL_KEYS, 'unknown key %r' % name
         rv = self.sreg.get(name)
@@ -180,10 +187,10 @@ class OpenIDResponse(object):
     the :meth:`~OpenID.try_login` function.
     """
 
-    def __init__(self, resp):
+    def __init__(self, resp, extensions):
         #: the openid the user used for sign in
         self.identity_url = resp.identity_url
-        lookup = RegLookup(resp)
+        lookup = RegLookup(resp, extensions)
 
         #: the full name of the user
         self.fullname = lookup.get_combined('fullname', FULL_NAME_URIS)
@@ -278,6 +285,10 @@ class OpenIDResponse(object):
         #: URL to profile image as string
         self.image = lookup.get('image')
 
+        #: Hash of the response object from the OpenID Extension by the
+        # OpenID Extension class name
+        self.extensions = lookup.ext
+
 
 class OpenID(object):
     """Simple helper class for OpenID auth.  Has to be created in advance
@@ -312,10 +323,11 @@ class OpenID(object):
                               if the HTTP referrer is unreliable.  By
                               default the user is redirected back to the
                               application's index in that case.
+    :param extension_responses: a list of OpenID Extensions Response class.
     """
 
     def __init__(self, app=None, fs_store_path=None, store_factory=None,
-                 fallback_endpoint=None):
+                 fallback_endpoint=None, extension_responses=None, safe_roots=None):
         # backwards compatibility support
         if isinstance(app, basestring):
             from warnings import warn
@@ -338,6 +350,13 @@ class OpenID(object):
         self.store_factory = store_factory
         self.after_login_func = None
         self.fallback_endpoint = fallback_endpoint
+        if not extension_responses:
+            extension_responses = []
+        self.extension_responses = extension_responses
+        if isinstance(safe_roots, basestring):
+            self.safe_roots = [safe_roots]
+        else:
+            self.safe_roots = safe_roots
 
     def init_app(self, app):
         """This callback can be used to initialize an application for the
@@ -382,12 +401,25 @@ class OpenID(object):
         always return a valid URL.
         """
         return (
-            request.values.get('next') or
-            request.referrer or
-            (self.fallback_endpoint and url_for(self.fallback_endpoint)) or
+            self.check_safe_root(request.values.get('next')) or
+            self.check_safe_root(request.referrer) or
+            (self.fallback_endpoint and self.check_safe_root(url_for(self.fallback_endpoint))) or
             request.url_root
         )
 
+    def check_safe_root(self, url):
+        if url is None:
+            return None
+        if self.safe_roots is None:
+            return url
+        if url.startswith(request.url_root) or url.startswith('/'):
+            # A URL inside the same app is deemed to always be safe
+            return url
+        for safe_root in self.safe_roots:
+            if url.startswith(safe_root):
+                return url
+        return None
+
     def get_current_url(self):
         """the current URL + next."""
         return request.base_url + '?next=' + url_quote(self.get_next_url())
@@ -399,16 +431,18 @@ class OpenID(object):
         """
         return self.get_current_url() + '&openid_complete=yes'
 
-    def attach_reg_info(self, auth_request, keys):
+    def attach_reg_info(self, auth_request, keys, optional_keys):
         """Attaches sreg and ax requests to the auth request.
 
         :internal:
         """
-        keys = set(keys)
+        keys = set(keys or [])
+        optional_keys = set(optional_keys or [])
         sreg_keys = list(SREG_KEYS & keys)
-        auth_request.addExtension(SRegRequest(required=sreg_keys))
+        sreg_optional_keys = list(SREG_KEYS & optional_keys)
+        auth_request.addExtension(SRegRequest(required=sreg_keys, optional=sreg_optional_keys))
         ax_req = ax.FetchRequest()
-        for key in keys:
+        for key in (keys | optional_keys):
             for uri in AX_MAPPING.get(key, ()):
                 ax_req.add(ax.AttrInfo(uri, required=key in REQUIRED_KEYS))
         auth_request.addExtension(ax_req)
@@ -445,10 +479,12 @@ class OpenID(object):
             if request.args.get('openid_complete') != u'yes':
                 return f(*args, **kwargs)
             consumer = Consumer(SessionWrapper(self), self.store_factory())
-            openid_response = consumer.complete(request.args.to_dict(),
-                                                self.get_current_url())
+            args = request.args.to_dict()
+            args.update(request.form.to_dict())
+            openid_response = consumer.complete(args, self.get_current_url())
             if openid_response.status == SUCCESS:
-                return self.after_login_func(OpenIDResponse(openid_response))
+                return self.after_login_func(OpenIDResponse(
+                    openid_response, self.extension_responses))
             elif openid_response.status == CANCEL:
                 self.signal_error(u'The request was cancelled')
                 return redirect(self.get_current_url())
@@ -456,26 +492,40 @@ class OpenID(object):
             return redirect(self.get_current_url())
         return decorated
 
-    def try_login(self, identity_url, ask_for=None):
+    def try_login(self, identity_url, ask_for=None, ask_for_optional=None, extensions=None):
         """This tries to login with the given identity URL.  This function
-        must be called from the login_handler.  The `ask_for` parameter can
-        be a set of values to be asked from the openid provider.
+        must be called from the login_handler.  The `ask_for` and 
+        `ask_for_optional`parameter can be a set of values to be asked
+        from the openid provider, where keys in `ask_for` are marked as
+        required, and keys in `ask_for_optional` are marked as optional.
 
-        The following strings can be used in the `ask_for` parameter:
+        The following strings can be used in the `ask_for` and
+            `ask_for_optional` parameters:
         ``aim``, ``blog``, ``country``, ``dob`` (date of birth), ``email``,
         ``fullname``, ``gender``, ``icq``, ``image``, ``jabber``, ``language``,
         ``msn``, ``nickname``, ``phone``, ``postcode``, ``skype``,
         ``timezone``, ``website``, ``yahoo``
+
+        `extensions` can be a list of instances of OpenID extension requests
+        that should be passed on with the request. If you use this, please make
+        sure to pass the Response classes of these extensions when initializing
+        OpenID.
         """
         if ask_for and __debug__:
             for key in ask_for:
                 if key not in ALL_KEYS:
                     raise ValueError('invalid key %r' % key)
+            for key in ask_for_optional:
+                if key not in ALL_KEYS:
+                    raise ValueError('invalid optional key %r' % key)
         try:
             consumer = Consumer(SessionWrapper(self), self.store_factory())
             auth_request = consumer.begin(identity_url)
-            if ask_for:
-                self.attach_reg_info(auth_request, ask_for)
+            if ask_for or ask_for_optional:
+                self.attach_reg_info(auth_request, ask_for, ask_for_optional)
+            if extensions:
+                for extension in extensions:
+                    auth_request.addExtension(extension)
         except discover.DiscoveryFailure:
             self.signal_error(u'The OpenID was invalid')
             return redirect(self.get_current_url())
